@@ -9,7 +9,7 @@
 #define	MUTEX_EXIT(x)		spinunlock(x)
 #define	READ_ENTER(x)		lock_read(x)
 #define	WRITE_ENTER(x)		lock_write(x)
-#define	RW_DOWNGRADE(x)		lock_write_to_read(x0
+#define	RW_DOWNGRADE(x)		lock_write_to_read(x)
 #define	RW_EXIT(x)		lock_done(x)
 #define	KMALLOC(v,t,z,w)	do { MALLOC(v, t, z, M_IOSYS, w); \
 				     if ((v) != NULL) \
