@@ -7,7 +7,7 @@
  * Use is subject to license terms.
  */
 
-#pragma ident "@(#)$Id: qif.c,v 1.43 2004/01/15 11:24:01 darrenr Exp $"
+#pragma ident "@(#)$Id: qif.c,v 1.47 2004/10/28 13:21:32 darrenr Exp $"
 
 #include <sys/systm.h>
 #include <sys/types.h>
@@ -33,6 +33,7 @@
 #include <sys/dlpi.h>
 #include <sys/stropts.h>
 #include <sys/sockio.h>
+#include <sys/ethernet.h>
 #include <net/if.h>
 #if SOLARIS2 >= 6
 # include <net/if_types.h>
@@ -263,7 +264,6 @@ qif_ire_walker(ire, arg)
 /* Function:    qif_attach                                                  */
 /* Returns:     int  - 0 == success, -1 == error in attaching qif_t to q    */
 /* Parameters:  rq(I) - pointer to STREAMS read queue                       */
-/* Write Lock:  pfil_rw                                                     */
 /*                                                                          */
 /* Attempt to bind a qif_t structure to a specific interface given the      */
 /* queue pointer.  Assumes the queue already has a qif_t structure tagged   */
@@ -277,6 +277,7 @@ qif_attach(rq)
 	qif_t *qif;
 	ill_t *ill;
 
+	WRITE_ENTER(&pfil_rw);
 	/*
 	 * Can we map the queue to a specific ill?  If not, go no futher, we
 	 * are only interested in being associated with queues that we can
@@ -288,6 +289,7 @@ qif_attach(rq)
 			cmn_err(CE_NOTE,
 				"PFIL: cannot find interface for rq %p",
 				(void *)rq);
+		RW_EXIT(&pfil_rw);
 		return -1;
 	}
 
@@ -299,7 +301,7 @@ qif_attach(rq)
 	    (hdrsizes[ill->ill_type][0] == ill->ill_type))
 		qif->qf_hl = hdrsizes[ill->ill_type][1];
 
-	if (qif->qf_hl == 0) {
+	if (qif->qf_hl == 0 && ill->ill_type != IFT_OTHER) {
 		cmn_err(CE_WARN,
 			"!Unknown layer 2 header size for %s type %d sap %x\n",
 			qif->qf_name, ill->ill_type, ill->ill_sap);
@@ -312,8 +314,10 @@ qif_attach(rq)
 	 * all the information has been set with qf_bound finally set to 1
 	 * after that.
 	 */
-	if (qif->qf_bound == 1)
+	if (qif->qf_bound == 1) {
+		RW_EXIT(&pfil_rw);
 		return 0;
+	}
 
 	qif->qf_sap = ill->ill_sap;
 	qif->qf_ppa = ill->ill_ppa;
@@ -324,7 +328,8 @@ qif_attach(rq)
 	qif->qf_name[sizeof(qif->qf_name) - 1] = '\0';
 	qif->qf_ill = ill;
 	qif->qf_bound = 1;
-	qif_ipmp_syncslave(qif);
+	qif_ipmp_syncslave(qif, qif->qf_sap);
+	RW_EXIT(&pfil_rw);
 
 	READ_ENTER(&pfh_sync.ph_lock);
 
@@ -390,7 +395,6 @@ qif_new(q, mflags)
 /* Function:    qif_delete                                                  */
 /* Returns:     void                                                        */
 /* Parameters:  None.                                                       */
-/* Write Locks: pfil_rw                                                     */
 /*                                                                          */
 /* Remove a qif structure from the list of recognised qif's.                */
 /*                                                                          */
@@ -415,6 +419,7 @@ queue_t *q;
 	if (qif == NULL)
 		return;
 
+	WRITE_ENTER(&pfil_rw);
 	PT_ENTER_WRITE(&qif->qf_ptl);
 
 	if (qif->qf_bound == 1 && qif_verbose > 0)
@@ -428,6 +433,7 @@ queue_t *q;
 			break;
 		}
 	PT_EXIT_WRITE(&qif->qf_ptl);
+	RW_EXIT(&pfil_rw);
 
 	if (qif->qf_ill) {
 		READ_ENTER(&pfh_sync.ph_lock);
@@ -575,8 +581,20 @@ qif_t *qif_walk(qif_t **qfp)
 void qif_ipmp_update(char *ipmpconf)
 {
 	qif_t *qif, *qf;
+	int len, sap;
 	char *s;
-	int len;
+
+	sap = ETHERTYPE_IP;
+	if (!strncmp(ipmpconf, "v4:", 3)) {
+		ipmpconf += 3;
+	} else if (!strncmp(ipmpconf, "v6:", 3)) {
+#if SOLARIS2 >= 8
+		sap = IP6_DL_SAP;
+		ipmpconf += 3;
+#else
+		return;
+#endif
+	}
 
 	s = strchr(ipmpconf, '=');
 	if (s != NULL) {
@@ -607,6 +625,7 @@ void qif_ipmp_update(char *ipmpconf)
 		qif->qf_next = qif_head;
 		qif_head = qif;
 
+		qif->qf_sap = sap;
 		qif->qf_flags |= QF_IPMP;
 		qif->qf_qifsz = len;
 		qif->qf_members = (char *)qif + sizeof(*qif);
@@ -615,11 +634,10 @@ void qif_ipmp_update(char *ipmpconf)
 		KMFREE(qif, len);
 		qif = qf;
 	}
-	RW_DOWNGRADE(&pfil_rw);
 
 	strcpy(qif->qf_members, s);
 
-	qif_ipmp_syncmaster(qif);
+	qif_ipmp_syncmaster(qif, sap);
 
 	RW_EXIT(&pfil_rw);
 }
@@ -676,7 +694,7 @@ void qif_ipmp_delete(char *qifname)
 /* rechecked in case they have become excluded. This function should be     */
 /* called for any new IPMP qif's created or when an IPMP qif changes.       */
 /* ------------------------------------------------------------------------ */
-void qif_ipmp_syncmaster(qif_t *updated)
+void qif_ipmp_syncmaster(qif_t *updated, const int sap)
 {
 	char *s, *t;
 	qif_t *qf;
@@ -684,6 +702,8 @@ void qif_ipmp_syncmaster(qif_t *updated)
 	for (qf = qif_head; qf != NULL; qf = qf->qf_next)  {
 		if ((qf->qf_flags & QF_IPMP) != 0)
 			continue;
+		if (qf->qf_sap != sap)
+			continue;
 		if (qf->qf_ipmp == updated)
 			qf->qf_ipmp = NULL;
 		for (s = updated->qf_members; s != NULL; ) {
@@ -709,7 +729,7 @@ void qif_ipmp_syncmaster(qif_t *updated)
 /* Check through the list of qif's to see if there is an IPMP with a member */
 /* list that includes the one named by target.                              */
 /* ------------------------------------------------------------------------ */
-void qif_ipmp_syncslave(qif_t *target)
+void qif_ipmp_syncslave(qif_t *target, const int sap)
 {
 	char *s, *t;
 	qif_t *qf;
@@ -723,6 +743,8 @@ void qif_ipmp_syncslave(qif_t *target)
 	for (qf = qif_head; qf != NULL; qf = qf->qf_next)  {
 		if ((qf->qf_flags & QF_IPMP) == 0)
 			continue;
+		if (qf->qf_sap != sap)
+			continue;
 		for (s = qf->qf_members; s != NULL; ) {
 			t = strchr(s, ',');
 			if (t != NULL)
@@ -737,3 +759,63 @@ void qif_ipmp_syncslave(qif_t *target)
 		}
 	}
 }
+
+
+/* ------------------------------------------------------------------------ */
+/* Function:    qif_hl_set                                                  */
+/* Returns:     void                                                        */
+/* Parameters:  ipmpconf(I) - string with header length setting for NIC     */
+/*                                                                          */
+/* For NICs that we cannot automatically determine the MAC header length of */
+/* we provide a manual crook to achieve that with.  The input syntax for    */
+/* the string is "[v4:|v6:]<ifname>=<length>"                               */
+/* ------------------------------------------------------------------------ */
+void qif_hl_set(char *ipmpconf)
+{
+	qif_t *qif, *qf;
+	int len, sap;
+	char *s;
+
+	sap = ETHERTYPE_IP;
+	if (!strncmp(ipmpconf, "v4:", 3)) {
+		ipmpconf += 3;
+	} else if (!strncmp(ipmpconf, "v6:", 3)) {
+#if SOLARIS2 >= 8
+		sap = IP6_DL_SAP;
+		ipmpconf += 3;
+#else
+		return;
+#endif
+	}
+
+	s = strchr(ipmpconf, '=');
+	if (s != NULL) {
+		if (*(s + 1) == '\0')
+			*s = '\0';
+		else
+			*s++ = '\0';
+	}
+	if (s == NULL || *s == NULL)
+		return;
+
+	READ_ENTER(&pfil_rw);
+	for (qf = qif_head; qf; qf = qf->qf_next) 
+		if (strcmp(qf->qf_name, ipmpconf) == 0)
+			break;
+
+	if (qf != NULL) {
+		int hl = 0;
+
+		for (; *s != '\0'; s++) {
+			char c = *s;
+
+			if (c < '0' || c > '9')
+				return;
+			hl *= 10;
+			hl += c - '0'; 
+		}
+		qf->qf_hl = hl;
+	}
+
+	RW_EXIT(&pfil_rw);
+}
