@@ -275,7 +275,7 @@ void pfilwput_ioctl(queue_t *q, mblk_t *mp)
 			for (; pfh; pfh = pfh->pfil_next)
 				if (pfh->pfil_func)
 					(void) (*pfh->pfil_func)(NULL, 0,
-								 qif->qf_ill,
+								 qif,
 								 0, qif, NULL);
 			RW_EXIT(&pfh_sync.ph_lock);
 		}
